/*
 * OpenGL window implementation.
 */

#include <glibmm/error.h>
#include <cstdlib>
#include <GL/gl.h>
#include <GL/glu.h>
#include <sstream>
#include <gtkmm/messagedialog.h>
#include <gdk/gdkkeysyms.h>
#include <gdkmm/cursor.h>

#include "interface.hpp"

/* -- GlDrawingArea default constructor. --------------------------------- */
Interface::GlDrawingArea::GlDrawingArea (): Gtk::DrawingArea () {
  /* OpenGL configuration. */
  Glib::RefPtr<Gdk::GL::Config> configuration;

  /* Trying double buffer mode. */
  configuration = Gdk::GL::Config::create(Gdk::GL::MODE_RGBA   |
                                          Gdk::GL::MODE_DEPTH  |
                                          Gdk::GL::MODE_DOUBLE);
  if (!configuration) {
    /* Warning message. */
    std::stringstream message;
    message << "Impossible to configure double buffered OpenGL mode, "
            << "trying single buffered mode.\n";
    g_warning(message.str().c_str());
    /* Trying to configure single buffered mode. */
    configuration = Gdk::GL::Config::create(Gdk::GL::MODE_RGBA  |
                                            Gdk::GL::MODE_DEPTH);
    if (!configuration) throw Errors::GLConfigurationError ();
  }

  /* Gives OpenGL capability to the widget. */
  set_gl_capability(configuration);
}

/* -- GlDrawingArea destructor. ------------------------------------------ */
Interface::GlDrawingArea::~GlDrawingArea () {
}

/* -- OpenGL context configuration. -------------------------------------- */
void Interface::GlDrawingArea::on_realize () {
  /* First standard initialisation. */
  Gtk::Widget::on_realize();

  if (!create_gl_context()) throw Errors::ContextCreationError ();
  /* OpenGL window. */
  Glib::RefPtr<Gdk::GL::Window> glWindow = get_gl_window();
  if (!glWindow) throw Errors::GettingContextError ();

  glWindow->gl_begin(get_gl_context());

    /* Great shading. */
    glShadeModel(GL_SMOOTH);

    /* Black foreground. */
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

    /* Initialising depth buffer. */
    glClearDepth(1.0f);
    /* Enabling depth test. */
    glEnable(GL_DEPTH_TEST);
    /* Setting depth test mode. */
    glDepthFunc(GL_LEQUAL);

    /* Nice perspective computation. */
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

  glWindow->gl_end();
}

/* -- When displaying GlDrawingArea. ------------------------------------- */
bool Interface::GlDrawingArea::on_expose_event (GdkEventExpose* event) {
  /* OpenGL window. */
  Glib::RefPtr<Gdk::GL::Window> glWindow = get_gl_window();
  if (!glWindow)  throw Errors::GettingContextError ();

  glWindow->gl_begin(get_gl_context());

    /* Clear screen and Z-Buffer (Depth Buffer: contains each pixel
       depth). */
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    /* Camera re-positionning to screen center. */
    glLoadIdentity();

    /* Moves 1.5 units left and 6.0 units inside the screen. */
    glTranslatef(-1.5f,0.0f,-6.0f);

    /* Drawing a triangle. */
    glBegin(GL_TRIANGLES);
      /* Upper vertex. */
      glVertex3f( 0.0f, 1.0f, 0.0f);
      /* Lower left vertex. */
      glVertex3f(-1.0f,-1.0f, 0.0f);
      /* Lower right vertex. */
      glVertex3f( 1.0f,-1.0f, 0.0f);
    /* End of triangle drawing. */
    glEnd();

    /* Moves 3 units right. */
    glTranslatef(3.0f,0.0f,0.0f);

    /* Dessin d'un quadrilatère. */
    glBegin(GL_QUADS);
      /* Upper left. */
      glVertex3f(-1.0f, 1.0f, 0.0f);
      /* Upper right. */
      glVertex3f( 1.0f, 1.0f, 0.0f);
      /* Lower right. */
      glVertex3f( 1.0f,-1.0f, 0.0f);
      /* Lower left. */
      glVertex3f(-1.0f,-1.0f, 0.0f);
    glEnd();
  
  glWindow->gl_end();
  glWindow->swap_buffers();

  /* Standard drawing procedure. */
  return Gtk::Widget::on_expose_event(event);
}

/* -- When resizing MainWindow. ------------------------------------------ */
bool Interface::GlDrawingArea::on_configure_event (GdkEventConfigure* event) {
  /* OpenGL window. */
  Glib::RefPtr<Gdk::GL::Window> glWindow = get_gl_window();
  if (!glWindow)  throw Errors::GettingContextError ();

  /* Rendering area width. */
  const GLsizei width = static_cast<GLsizei>(get_width());
  /* Rendering area height. */
  const GLsizei height = static_cast<GLsizei>(get_height());

  glWindow->gl_begin(get_gl_context());

    /* Setting rendering area size. */
    glViewport(0, 0, width, height);
    /* Picking projection matrix. */
    glMatrixMode(GL_PROJECTION);
    /* Clearing projection matrix. */
    glLoadIdentity();

    /* Perspective ratio for the rendering area. */
    gluPerspective(45.0f, static_cast<GLfloat>(width)
                            / static_cast<GLfloat>(height), 0.1f,100.0f);
    /* Picking modelling matrix. */
    glMatrixMode(GL_MODELVIEW);
    /* Clearing modelling matrix. */
    glLoadIdentity();

  glWindow->gl_end();

  /* Standard part of resizing. */
  return Gtk::Widget::on_configure_event(event);
}

/* -- MainWindow default constructor. ----------------------------- */
Interface::MainWindow::MainWindow (): Gtk::Window (), renderingArea () {
  set_title("Lesson 2");
  set_default_size(640, 480);
  /* Minimum size, to avoid zero devide. */
  set_size_request(1, 1);

  add(renderingArea);

  add_events(Gdk::KEY_RELEASE_MASK);

  signal_key_release_event().connect(sigc::mem_fun(*this,
                                                  &MainWindow::on_key_release));
  signal_window_state_event().connect(sigc::mem_fun(*this,
                                           &MainWindow::on_window_state_event));

  /* Dialogue box for full-screen mode. */
  Gtk::MessageDialog pleinEcran (*this, "Switching in full-screen mode?",
                                 false, Gtk::MESSAGE_QUESTION,
                                 Gtk::BUTTONS_YES_NO);
  /* User’s choice. */
  int choice = pleinEcran.run();

  if (choice == Gtk::RESPONSE_YES) {
    inFullScreen = true;
    show();
    fullscreen();

    /* Creates an empty cursor. */
    const Gdk::Cursor cursor (Gdk::Display::get_default(),
                              Gdk::Pixbuf::create(Gdk::COLORSPACE_RGB, true,
                                                  8, 1, 1), 0, 0);
    /* Display empty cursor, which result in hidding cursor. */
    get_window()->set_cursor(cursor);
  }
  else {
    set_border_width(2);
    inFullScreen = false;
  }

    show_all_children();
}

/* -- MainWindow destructor. --------------------------------------------- */
Interface::MainWindow::~MainWindow () {
}

/* -- Handling keyboard events. ------------------------------------------ */
bool Interface::MainWindow::on_key_release (GdkEventKey* evenement) {
  switch (evenement->keyval) {
    case GDK_KEY_Escape:
      hide();
      break;
    case GDK_KEY_F1:
      if (inFullScreen) {
        inFullScreen = false;
        set_border_width(2);
        unfullscreen();
        /* Back to default cursor, which results in showing it back. */
        get_window()->set_cursor();
      }
      else {
        inFullScreen = true;
        set_border_width(0);
        fullscreen();
        /* Creates an empty cursor. */
        const Gdk::Cursor cursor (Gdk::Display::get_default(),
                                  Gdk::Pixbuf::create(Gdk::COLORSPACE_RGB,
                                                      true, 8, 1, 1), 0, 0);
        /* Display empty cursor, which results in hidding it. */
        get_window()->set_cursor(cursor);
      }
  }

  return true;
}

/* -- When switching in full-screen mode --------------------------------- */
bool Interface::MainWindow::on_window_state_event(GdkEventWindowState* pEven) {
  /* Standard window updating result. */
  const bool result = Gtk::Widget::on_window_state_event(pEven);

  if (result) {
    /* Make sure screen is in intended mode. */
    if (inFullScreen
        && (pEven->changed_mask != GDK_WINDOW_STATE_FULLSCREEN)) {
      set_border_width(2);
      inFullScreen = false;
      unfullscreen();

      /* Error message box. */
      Gtk::MessageDialog pleinEcran (*this,
                                  "Impossible to switch in full-screen mode",
                                     false, Gtk::MESSAGE_QUESTION,
                                     Gtk::BUTTONS_YES_NO);
      pleinEcran.set_secondary_text("Keep on with windowed mode?");
      /* Choix de l’utilisateur. */
      int choice = pleinEcran.run();
      if (choice == Gtk::RESPONSE_NO) return false;
    }
  }

  return result;
}
